@@ -1,14 +1,16 @@
 import os
 import sys
+import glob
 import json
 import time
 import tempfile
 import importlib
 from copy import deepcopy
-from typing import Optional, Dict, List, Callable, Union
+from typing import Optional, Dict, List, Callable, Union, Tuple
 from datetime import datetime
 
 import numpy as np
+import pandas as pd
 
 from hpogrid.core.defaults import *
 from hpogrid.core import config_manager
@@ -137,6 +139,10 @@ class HPOTask(AbstractObject):
     @property
     def best_config(self) -> Dict:
         return self._best_config
+    
+    @property
+    def best_metric_score(self) -> Dict:
+        return self._best_metric_score
 
     @property
     def hyperparameters(self):
@@ -232,15 +238,17 @@ class HPOTask(AbstractObject):
         # search space
         self.set_search_space()
         # parameters for logging
+        self.outdir = "./"
         self._df = None
         self._start_datetime = None
         self._end_datetime = None
         self._total_time = None
         self._best_config = None
+        self._best_metric_score = None
         self._start_timestamp = None
         self.idds_job = False
         self.search_points = None
-        self.summary_fname = kGridSiteMetadataFileName
+        self._df_restore = None
         
     def set_model(self, model:Optional[Callable]=None):
         self._model = model
@@ -248,7 +256,7 @@ class HPOTask(AbstractObject):
     def set_search_space(self, search_space:Optional[Dict]=None):
         if search_space is None:
             self._search_space = None
-            self._hyperparameters = {}
+            self._hyperparameters = []
         elif isinstance(search_space, dict):
             self._search_space = deepcopy(search_space)
             self._hyperparameters = list(self.search_space.keys())
@@ -322,35 +330,35 @@ class HPOTask(AbstractObject):
             raise ValueError(f'Unrecognized search algorithm: {algorithm}')
 
     @staticmethod
-    def get_algorithm(name:str, metric:str, mode:str, base_search_space:Dict,
-                      max_concurrent=None, **args):
+    def get_algorithm(name:str, metric:str, mode:str, search_space:Dict,
+                      max_concurrent:int=None, restore_points:Optional[Tuple]=None,
+                      **kwargs):
         _name = AlgorithmType.parse(name)
         if _name == AlgorithmType.AX:
             from hpogrid.algorithms import AxAlgorithm
-            algorithm = AxAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = AxAlgorithm
         elif _name == AlgorithmType.BOHB:
             from hpogrid.algorithms import BOHBAlgorithm
-            algorithm = BOHBAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = BOHBAlgorithm
         elif _name == AlgorithmType.HYPEROPT:
             from hpogrid.algorithms import HyperOptAlgorithm
-            algorithm = HyperOptAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = HyperOptAlgorithm
         elif _name == AlgorithmType.SKOPT:
             from hpogrid.algorithms import SkOptAlgorithm
-            algorithm = SkOptAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = SkOptAlgorithm
         elif _name == AlgorithmType.NEVERGRAD:
             from hpogrid.algorithms import NeverGradAlgorithm
-            algorithm = NeverGradAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = NeverGradAlgorithm
         elif _name == AlgorithmType.OPTUNA:
             from hpogrid.algorithms import OptunaAlgorithm
-            algorithm = OptunaAlgorithm().create(metric, mode, base_search_space, **args)
+            algo_cls = OptunaAlgorithm
         elif _name == AlgorithmType.TUNE:
-            algorithm = None
+            return None
         else:
             raise ValueError(f'Unrecognized search algorithm: {name}')
-        # limit max concurrency
-        if algorithm and max_concurrent:
-            from ray.tune.suggest import ConcurrencyLimiter
-            algorithm = ConcurrencyLimiter(algorithm, max_concurrent=max_concurrent) 
+        algo_wrapper = algo_cls(search_space=search_space, metric=metric, mode=mode)
+        algorithm = algo_wrapper.create(max_concurrent=max_concurrent, 
+                                        restore_points=restore_points, **kwargs)
         return algorithm
 
     def load_model(self, script_name:str, model_name:str, scripts_path:Optional[str]=None):
@@ -369,7 +377,31 @@ class HPOTask(AbstractObject):
         self.stdout.info('INFO: Loaded module {}'.format(model.__name__))
         return model
 
-    def create_metadata(self, df) -> Dict:
+    def _process_tune_df(self, df:"pandas.DataFrame"):
+        rename_cols = { 'config/{}'.format(hp): hp for hp in self.hyperparameters}
+        rename_cols['time_total_s'] = 'time_s'
+        df = df.rename(columns=rename_cols)
+        cols_to_save = ['time_s'] + self.hyperparameters
+        cols_to_save.append(self.metric)
+        if self.extra_metrics:
+            cols_to_save += self.extra_metrics
+        cols_to_save = list(set(cols_to_save))
+        df = df.filter(cols_to_save, axis=1)
+        return df
+    
+    def get_processed_df(self, df:Optional["pandas.DataFrame"]=None):
+        if df is not None:
+            df = self._process_tune_df(df)
+            if self._df_restore is not None:
+                df = pd.concat([df, self._df_restore], ignore_index=True)
+        elif self._df_restore is not None:
+            df = self._df_restore
+        else:
+            raise RuntimeError("no metadata to restore")
+        return df
+        
+    def create_metadata(self, df:Optional["pandas.DataFrame"]=None) -> Dict:
+        # if None: use cached result only
 
         summary = {
             'project_name' : self.project_name,
@@ -380,19 +412,13 @@ class HPOTask(AbstractObject):
             'hyperparameters': self.hyperparameters,
             'metric': self.metric,
             'mode' : self.mode,
-            'best_config' : self.best_config, 
+            'best_config': self.best_config,
+            'best_metric_score': self.best_metric_score
         }
+         
+        df = self.get_processed_df(df)
+        df = df.transpose()
         
-        rename_cols = { 'config/{}'.format(hp): hp for hp in self.hyperparameters}
-        rename_cols['time_total_s'] = 'time_s'
-        
-        df = df.rename(columns=rename_cols)
-        cols_to_save = ['time_s'] + self.hyperparameters
-        if 'metric' in summary:
-            cols_to_save.append(summary['metric'])
-        if self.extra_metrics:
-            cols_to_save += self.extra_metrics
-        df = df.filter(cols_to_save, axis=1).transpose()
         summary['result'] = df.to_dict()
         
         return summary
@@ -415,8 +441,16 @@ class HPOTask(AbstractObject):
         from hpogrid.utils.stylus import create_formatted_dict
         text = create_formatted_dict(self.export_setup(), indexed=False, columns=["Attribute","Value"])
         self.stdout.info(text)
+        
+    def get_all_metrics(self):
+        all_metrics = [self.metric]
+        if self.extra_metrics:
+            for extra_metric in self.extra_metrics:
+                if extra_metric not in all_metrics:
+                    all_metrics.append(extra_metric)
+        return all_metrics
     
-    def run(self) -> None:
+    def run(self, outdir:Optional[str]=None, cache:bool=True) -> None:
             
         self._start_timestamp = start = time.time()
         self._start_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
@@ -430,24 +464,7 @@ class HPOTask(AbstractObject):
             tune_config_space = {}
 
         tune_config_space.update(self.model_param)
-
-        # get search algorithm
-        algorithm = self.get_algorithm(
-            self.algorithm,
-            self.metric,
-            self.mode,
-            self.search_space,
-            self.max_concurrent,
-            **self.algorithm_param)
-        
-        # get trial scheduler
-        scheduler = self.get_scheduler(
-            self.scheduler,
-            self.metric,
-            self.mode,
-            self.search_space,
-            **self.scheduler_param)
-        
+                
         # load model from path
         if (not self.model):
             model = self.load_model(self.model_script, self.model_name, self.scripts_path)
@@ -456,104 +473,257 @@ class HPOTask(AbstractObject):
             
         self.print_resource_info()
         
-        callbacks = None
-        #loggers = self.get_loggers()
+        if outdir is None:
+            outdir = self.outdir
+        
+        metadata_path = os.path.join(outdir, kGridSiteMetadataFileName)
+        cache_dir     = os.path.join(outdir, "cache")
+        
+        if cache and os.path.exists(cache_dir):
+            self.stdout.info(f"INFO: Attempting to restore search results from \"{cache_dir}\"")
+            restore_points = self._get_restore_points(cache_dir, load_df=True)
+        else:
+            restore_points = None
+
+        if restore_points:
+            n_cached = len(restore_points[0])
+            n_restored = len([v for v in restore_points[1] if not helper.is_nan_or_inf(v)])
+            self.stdout.info(f"INFO: Found {n_cached} search results from cache")
+            self.stdout.info(f"INFO: Will restore {n_restored} valid search results")
+        else:
+            n_cached = 0
+            n_restored = 0
+
+        # get trial scheduler
+        scheduler = self.get_scheduler(
+            self.scheduler,
+            self.metric,
+            self.mode,
+            self.search_space,
+            **self.scheduler_param)
+ 
+        # get callbacks
+        callbacks = self.get_callbacks(cache_dir=cache_dir)
+        
+        # determine the number of search points to run accounting for restored points
+        if self.num_trials == -1:
+            num_samples = max(self.resource.get('gpu'), 1)
+        else:
+            num_samples = self.num_trials
+            
+        if (self.num_trials != -1) and (n_restored > 0):
+            num_samples_remain = num_samples - n_restored
+        else:
+            num_samples_remain = num_samples
+                
+        if not os.path.exists(outdir):
+            os.makedirs(outdir)
+        if not os.path.exists(cache_dir):
+            os.makedirs(cache_dir)
         
         # save something to prevent looping job
-        with open(kGridSiteMetadataFileName,'w') as output:
+        with open(metadata_path,'w') as output:
             json.dump({}, output)
-        
+
         temp_dir = tempfile.TemporaryDirectory()
         
-        import ray
-        from ray import tune
+        if not self.log_dir:
+            local_dir = os.path.join(temp_dir.name, 'log')
+        else:
+            local_dir = self.log_dir
 
-        # HPO starts
-        try:
-            # extract input dataset if it is a tar file
-            # (for grid jobs only)
-            extracted_files = []
-            if es.is_grid_job():
-                datadir = es.get_datadir()
-                extracted_files = helper.extract_tarball(datadir, datadir)
-            ray.init(include_dashboard=False)
-   
-            if not self.log_dir:
-                local_dir = os.path.join(temp_dir.name, 'log')
-            else:
-                local_dir = self.log_dir
-            if self.num_trials == -1:
-                num_samples = max(self.resource.get('gpu'), 1)
-            else:
-                num_samples = self.num_trials
-                
-            if AlgorithmType.parse(self.algorithm) == AlgorithmType.NEVERGRAD:
-                try:
-                    algo_str = f"{self.algorithm} ({algorithm.searcher._nevergrad_opt.name})"
-                except:
+        if num_samples_remain > 0:
+            
+            # get search algorithm
+            # move to here so that if jobs are already done we don't need to restore
+            algorithm = self.get_algorithm(
+                self.algorithm,
+                self.metric,
+                self.mode,
+                self.search_space,
+                self.max_concurrent,
+                restore_points=restore_points,
+                **self.algorithm_param)
+            
+            import ray
+            from ray import tune
+
+            # HPO starts
+            try:
+                # extract input dataset if it is a tar file
+                # (for grid jobs only)
+                extracted_files = []
+                if es.is_grid_job():
+                    datadir = es.get_datadir()
+                    extracted_files = helper.extract_tarball(datadir, datadir)
+                ray.init(include_dashboard=False)
+
+                if AlgorithmType.parse(self.algorithm) == AlgorithmType.NEVERGRAD:
+                    try:
+                        algo_str = f"{self.algorithm} ({algorithm.searcher._nevergrad_opt.name})"
+                    except:
+                        algo_str = f"{self.algorithm}"
+                else:
                     algo_str = f"{self.algorithm}"
-            else:
-                algo_str = f"{self.algorithm}"
                 
-            self.stdout.info("--------------------------------------------------------------------------------")
-            self.stdout.info(f"INFO: Begin hyperparameter optimization with the following settings")
-            self.stdout.info(f"Project Name            : {self.project_name}")
-            self.stdout.info(f"Search Algorithm        : {algo_str}")
-            self.stdout.info(f"Trial Scheduler         : {self.scheduler}")
-            self.stdout.info(f"Number of Search Points : {num_samples}")
-            self.stdout.info("--------------------------------------------------------------------------------")
-            cwd = os.getcwd()
-            analysis = tune.run(
-                model,
-                name=self.project_name,
-                scheduler=scheduler,
-                search_alg=algorithm,
-                config=tune_config_space,
-                num_samples=num_samples,
-                resources_per_trial=self.resource_per_trial,
-                verbose=self.verbose,
-                local_dir=local_dir,
-                callbacks=callbacks,
-                stop=self.stop)
-            os.chdir(cwd)
-        finally:
-            ray.shutdown()
-            # remove the extracted tar files
-            helper.remove_files(extracted_files)
+                if num_samples_remain != num_samples:
+                    sample_str = f"{num_samples} ({num_samples - num_samples_remain} restored)"
+                else:
+                    sample_str = f"{num_samples}"
+
+                self.stdout.info(f"INFO: Ray Tune log files will be saved to {local_dir}")
+                self.stdout.info("--------------------------------------------------------------------------------")
+                self.stdout.info(f"INFO: Begin hyperparameter optimization with the following settings")
+                self.stdout.info(f"\tProject Name            : {self.project_name}")
+                self.stdout.info(f"\tSearch Algorithm        : {algo_str}")
+                self.stdout.info(f"\tTrial Scheduler         : {self.scheduler}")
+                self.stdout.info(f"\tNumber of Search Points : {sample_str}")
+                self.stdout.info("--------------------------------------------------------------------------------")
+                cwd = os.getcwd()
+                analysis = tune.run(
+                    model,
+                    name=self.project_name,
+                    scheduler=scheduler,
+                    search_alg=algorithm,
+                    config=tune_config_space,
+                    num_samples=num_samples_remain,
+                    resources_per_trial=self.resource_per_trial,
+                    verbose=self.verbose,
+                    local_dir=local_dir,
+                    callbacks=callbacks,
+                    raise_on_failed_trial=False,
+                    stop=self.stop)
+                os.chdir(cwd)
+            finally:
+                ray.shutdown()
+                # remove the extracted tar files
+                helper.remove_files(extracted_files)
+            
+            if num_samples_remain == num_samples:
+                best_trial = analysis.get_best_trial(metric=self.metric, mode=self.mode)
+                if best_trial:
+                    self._best_config = best_trial.config
+                    self._best_metric_score = best_trial.metric_analysis[self.metric].get("last", None)
+                else:
+                    self._best_config = None
+                    self._best_metric_score = None
+                    
+            df = analysis.dataframe(metric=self.metric, mode=self.mode)
+        else:
+            self.stdout.info(f"INFO: Number of restored points({n_restored}) "
+                             f">= number of search points({num_samples}). "
+                             "Skip execution.")
+            df = None
+
+        if not self.best_config:
+            self._best_config, self._best_metric_score = self.get_best_trial_from_tune_df(df)
+            
         end = time.time()
         self._total_time = float(end - start)
         self._end_datetime = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
-
-        self._best_config = analysis.get_best_config(metric=self.metric, mode=self.mode)
-        
         self.stdout.info(f"Best config: {self.best_config}")
-        self.stdout.info(f"Time taken in seconds: {self.total_time}")
+        self.stdout.info(f"Best metric score: {self.best_metric_score}")
+        self.stdout.info(f"INFO: Time taken in seconds: {self.total_time}")
         
-        if self.log_dir:
-            self.stdout.info(f"INFO: Ray Tune log files are saved at {self.log_dir}")
-
-        df = analysis.dataframe(metric=self.metric, mode=self.mode)
-
         metadata = self.create_metadata(df)
-        
-        summary_output_path = helper.pretty_path(self.summary_fname)
         # save metadata
-        with open(summary_output_path, 'w') as output:            
+        with open(metadata_path, 'w') as output:            
             json.dump(metadata, output, cls=helper.NpEncoder, indent=2)
-            self.stdout.info(f'INFO: Job metadata is saved at {summary_output_path}')
+            self.stdout.info(f'INFO: Job metadata is saved as {metadata_path}')
         
         # save idds output
         if es.is_idds_job():
             from hpogrid.interface.idds.main import save_idds_output_from_metadata
             save_idds_output_from_metadata(metadata)
-
+            
         # reset environment
         es.reset()
         temp_dir.cleanup()
     
-    def get_loggers(self):
-        from ray.tune.logger import NoopLogger, MLFLowLogger
-        return [NoopLogger, MLFLowLogger]
+    def get_best_trial_from_tune_df(self, df:Optional["pandas.DataFrame"]=None, metric:Optional[str]=None,
+                                    mode:Optional[str]=None, filter_nan_and_inf:bool=True):
+        if metric is None:
+            metric = self.metric
+        if mode is None:
+            mode = self.mode
+        # validation
+        if mode not in ["min", "max"]:
+            raise ValueError("optimization mode must be either \"min\" or \"max\"")
+            
+        best_config = None
+        best_metric_score = None
+        
+        hyperparameters = self.hyperparameters
+        
+        results = self.get_processed_df(df).transpose().to_dict()
+        
+        for i, result in results.items():
+            
+            if metric not in result:
+                continue
+                
+            metric_score = result[metric]
+            if filter_nan_and_inf and helper.is_nan_or_inf(metric_score):
+                continue
+                
+            if not set(hyperparameters).issubset(set(result.keys())):
+                continue
+            config =  {hp:result[hp] for hp in hyperparameters}
+            
+            if best_metric_score is None:
+                best_metric_score = metric_score
+                best_config = config
+                continue
+
+            if (mode == "max") and (best_metric_score < metric_score):
+                best_metric_score = metric_score
+                best_config = config
+            elif (mode == "min") and (best_metric_score > metric_score):
+                best_metric_score = metric_score
+                best_config = config
+
+        if not best_config:
+            self.stdout.warning("WARNING: Could not find best trial. "
+                                "This can be due to invalid \"metric\" parameter "
+                                "or all trials ended up in invalid metric values (inf or nan).")
+        return best_config, best_metric_score
+        
+    def _get_restore_points(self, cache_dir:Optional[str]=None, load_df:bool=False):
+        if not cache_dir:
+            return None
+        filenames = glob.glob(os.path.join(cache_dir, "*.json"))
+        if not filenames:
+            return None
+
+        data = [json.load(open(filename)) for filename in filenames]
+        try:
+            for _data in data:
+                config = _data.pop("config")
+                for key, value in config.items():
+                    _data[f"config/{key}"] = value
+            df = pd.DataFrame(data)
+            df = self._process_tune_df(df)
+        except:
+            raise RuntimeError(f"invalid search results from {cache_dir}")
+            
+        if load_df:
+            self._df_restore = df
+            
+        hyperparams_values = df[self.hyperparameters].to_dict("records")
+        metric_values = df[self.metric].values
+        return (hyperparams_values, metric_values)
+    
+    def get_callbacks(self, cache_dir:Optional[str]=None):
+        #from ray.tune.logger import NoopLogger, MLFLowLogger
+        #return [NoopLogger, MLFLowLogger]
+        from hpogrid.interface.tune import TuneCallback
+        tune_callback = TuneCallback()
+        hyperparameters = self.hyperparameters
+        all_metrics = self.get_all_metrics()
+        tune_callback.custom_setup(hyperparameters=hyperparameters,
+                                   metrics=all_metrics,
+                                   cache_dir=cache_dir)
+        return [tune_callback]
     
     def set_run_mode(self, run_mode:str='local'):
         self._run_mode = run_mode
