@@ -22,18 +22,18 @@ class AxGenerator(Generator):
             minimize=minimize)
         return ax_client
 
-    def tell(self, point:Dict, value):
+    def _tell(self, searcher, point:Dict, value):
         value = self._to_metric_values(value)
-        _, trial_index = self.searcher.attach_trial(point)
-        metric_dict = { self.metric: (value, 0.0)}
-        self.searcher.complete_trial(
+        _, trial_index = searcher.attach_trial(point)
+        metric_dict = {self.metric: (value, 0.0)}
+        searcher.complete_trial(
             trial_index=trial_index, raw_data=metric_dict)
 
-    def ask(self, n_points:int =None):
+    def _ask(self, searcher, n_points:int =None):
         points = []
         for _ in range(n_points):
-            point, trial_index = self.searcher.get_next_trial()
-            trial = self.searcher._get_trial(trial_index=trial_index)
+            point, trial_index = searcher.get_next_trial()
+            trial = searcher._get_trial(trial_index=trial_index)
             trial.mark_abandoned()
             points.append(copy.deepcopy(point))
         return points
\ No newline at end of file
