@@ -9,22 +9,25 @@ class BOHBGenerator(Generator):
     kDefaultBudget = 100
 
     def get_searcher(self, search_space:Dict, metric:str, mode:str, **args):
-        from hpbandster.optimizers.config_generators.bohb import BOHB
+        
         search_space = BOHBSpace(search_space).get_search_space()
+        
+        from hpbandster.optimizers.config_generators.bohb import BOHB
         searcher = BOHB(search_space)
+        
         return searcher
 
-    def ask(self, n_points:int = None):
+    def _ask(self, searcher, n_points:int = None):
         points = []
         for _ in range(n_points):
-            point, info = self.searcher.get_config(None)
+            point, info = searcher.get_config(None)
             points.append(copy.deepcopy(point))
         return points
 
-    def tell(self, point:Dict, value:float):
+    def _tell(self, searcher, point:Dict, value:float):
         value = self._to_metric_values(value)
         result = BOHBResultWrapper(point, self.signature * value)
-        self.searcher.new_result(result)
+        searcher.new_result(result)
 
 class BOHBResultWrapper():
     
