@@ -9,10 +9,6 @@ class OptunaSpace(BaseSpace):
     def reset_space(self):
         self.search_space = {}
 
-    def get_search_space(self):
-        import nevergrad as ng
-        return ng.p.Instrumentation(**self.search_space)
-
     def append(self, space_value):
         self.search_space[space_value[0]] = space_value[1]
 
