@@ -1,9 +1,9 @@
+import math
 from past.utils import old_div
+
 import numpy as np
 
-from ray import tune
-
-from hpogrid.search_space.base_space import BaseSpace
+from .base_space import BaseSpace
 
 class TuneSpace(BaseSpace):
 
@@ -20,6 +20,7 @@ class TuneSpace(BaseSpace):
 
     def categorical(self, label, categories, grid_search = False):
         method = None
+        from ray import tune
         if grid_search:
             method = tune.grid_search(categories)
         else:
@@ -27,40 +28,50 @@ class TuneSpace(BaseSpace):
         return (label, method)
 
     def uniform(self, label, low, high):
+        from ray import tune
         return (label, tune.uniform(low, high))
     
     def uniformint(self, label, low, high):
+        from ray import tune
         return (label, tune.randint(low, high))
 
     def loguniform(self, label, low, high, base=10):
+        from ray import tune
         return (label, tune.loguniform(low, high, base))
 
     def qloguniform(self, label, low, high, base=10):
         logmin = np.log(low) / np.log(base)
         logmax = np.log(high) / np.log(base)
+        from ray import tune
         return (label, tune.sample_fram(lambda _: 
             np.round(old_div(base**(np.random.uniform(logmin, logmax)),q))*q))
 
     def quniform(self, label, low, high, q):
+        from ray import tune
         return (label, tune.sample_from(lambda _: 
             np.round(old_div(np.random.uniform(low, high),q))*q))
 
     def quniformint(self, label, low, high, q):
+        from ray import tune
         return (label, tune.sample_from(lambda _: 
             np.round(old_div(np.random.uniform(low, high),q))*q))
 
     def normal(self, mu, sigma):
+        from ray import tune
         return (label, tune.sample_from(lambda _: sigma*np.random.randn()+mu))
 
     def qnormal(self, mu, sigma, q):
+        from ray import tune
         return (label, tune.sample_from(lambda _: 
             np.round(old_div(sigma*np.random.randn()+mu,q))*q))
 
-    def lognormal(self, mu, sigma, base = np.e):
+    def lognormal(self, mu, sigma, base = math.e):
+        from ray import tune
         return (label, tune.sample_from(lambda _: 
             np.power(sigma*np.random.randn()+mu, base))) 
 
-    def qlognormal(self, mu, sigma, q, base = np.e):
+    def qlognormal(self, mu, sigma, q, base=math.e):
+        from ray import tune
         return (label, tune.sample_from(lambda _: 
             np.round(old_div(np.power(sigma*np.random.randn()+mu, base),q))*q ))         
 
