@@ -9,6 +9,8 @@ from contextlib import contextmanager
 from subprocess import Popen, PIPE
 import shlex
 
+import numpy as np
+
 def run_command(cmd):
     args = shlex.split(cmd, posix=False)
     with Popen(args, stdout=PIPE, bufsize=1, universal_newlines=True) as p:
@@ -53,6 +55,9 @@ def pretty_dirname(path:str):
         return path
     else:
         return os.path.dirname(path)
+    
+def is_nan_or_inf(value):
+    return np.isnan(value) or np.isinf(value)    
         
 @contextmanager
 def cd(newdir):
